# -*- coding: utf-8 -*-

import json
import sys

from elasticsearch import Elasticsearch
from elasticsearch.client.utils import query_params
from elasticsearch.exceptions import NotFoundError

from elasticmock.behaviour.server_failure import server_failure
from elasticmock.utilities import get_random_id, get_random_scroll_id
from elasticmock.utilities.decorator import for_all_methods
from elasticmock.fake_indices import FakeIndicesClient
from elasticmock.fake_cluster import FakeClusterClient

PY3 = sys.version_info[0] == 3
if PY3:
    unicode = str


@for_all_methods([server_failure])
class FakeElasticsearch(Elasticsearch):
    __documents_dict = None

    def __init__(self, hosts=None, transport_class=None, **kwargs):
        self.__documents_dict = {}
        self.__scrolls = {}

    @property
    def indices(self):
        return FakeIndicesClient(self)

    @property
    def cluster(self):
        return FakeClusterClient(self)

    @query_params()
    def ping(self, params=None, headers=None):
        return True

    @query_params()
    def info(self, params=None, headers=None):
        return {
            'status': 200,
            'cluster_name': 'elasticmock',
            'version':
                {
                    'lucene_version': '4.10.4',
                    'build_hash': '00f95f4ffca6de89d68b7ccaf80d148f1f70e4d4',
                    'number': '1.7.5',
                    'build_timestamp': '2016-02-02T09:55:30Z',
                    'build_snapshot': False
                },
            'name': 'Nightwatch',
            'tagline': 'You Know, for Search'
        }

    @query_params('consistency', 'op_type', 'parent', 'refresh', 'replication',
                  'routing', 'timeout', 'timestamp', 'ttl', 'version', 'version_type')
    def index(self, index, body, doc_type='_doc', id=None, params=None, headers=None):
        if index not in self.__documents_dict:
            self.__documents_dict[index] = list()

        version = 1

        if id is None:
            id = get_random_id()
        elif self.exists(index, doc_type, id, params=params):
            doc = self.get(index, id, doc_type, params=params)
            version = doc['_version'] + 1
            self.delete(index, doc_type, id)

        self.__documents_dict[index].append({
            '_type': doc_type,
            '_id': id,
            '_source': body,
            '_index': index,
            '_version': version
        })

        return {
            '_type': doc_type,
            '_id': id,
            'created': True,
            '_version': version,
            '_index': index
        }

    def allocation(self, index=None, params=None, headers=None):
        return {
            'shards': 1,
            'disk.indices': '260b',
            'disk.used': '47.3gb',
            'disk.avail': '43.4gb',
            'disk.total': '100.7gb',
            'disk.percent': '46',
            'host': '127.0.0.1',
            'ip': '127.0.0.1',
            'node': 'CSUXak2'
        }

    @query_params('consistency', 'op_type', 'parent', 'refresh', 'replication',
                  'routing', 'timeout', 'timestamp', 'ttl', 'version', 'version_type')
    def bulk(self,  body, index=None, doc_type=None, params=None, headers=None):
        version = 1
        items = []

        for line in body.splitlines():
            if len(line.strip()) > 0:
                line = json.loads(line)

                if 'index' in line:
                    index = line['index']['_index']
                    doc_type = line['index']['_type']

                    if index not in self.__documents_dict:
                        self.__documents_dict[index] = list()
                else:
                    document_id = get_random_id()

                    self.__documents_dict[index].append({
                        '_type': doc_type,
                        '_id': document_id,
                        '_source': line,
                        '_index': index,
                        '_version': version
                    })

                    items.append({'index': {
                        '_type': doc_type,
                        '_id': document_id,
                        '_index': index,
                        '_version': version,
                        'result': 'created',
                        'status': 201
                    }})

        return {
            'errors': False,
            'items': items
        }

    @query_params('parent', 'preference', 'realtime', 'refresh', 'routing')
    def exists(self, index, doc_type, id, params=None, headers=None):
        result = False
        if index in self.__documents_dict:
            for document in self.__documents_dict[index]:
                if document.get('_id') == id and document.get('_type') == doc_type:
                    result = True
                    break
        return result

    @query_params('_source', '_source_exclude', '_source_include', 'fields',
                  'parent', 'preference', 'realtime', 'refresh', 'routing', 'version',
                  'version_type')
    def get(self, index, id, doc_type='_all', params=None, headers=None):
        result = None
        if index in self.__documents_dict:
            for document in self.__documents_dict[index]:
                if document.get('_id') == id:
                    if doc_type == '_all':
                        result = document
                        break
                    else:
                        if document.get('_type') == doc_type:
                            result = document
                            break

        if result:
            result['found'] = True
        else:
            error_data = {
                '_index': index,
                '_type': doc_type,
                '_id': id,
                'found': False
            }
            raise NotFoundError(404, json.dumps(error_data))

        return result

    @query_params('_source', '_source_exclude', '_source_include', 'parent',
                  'preference', 'realtime', 'refresh', 'routing', 'version',
                  'version_type')
    def get_source(self, index, doc_type, id, params=None, headers=None):
        document = self.get(index=index, doc_type=doc_type, id=id, params=params)
        return document.get('_source')

    @query_params('_source', '_source_exclude', '_source_include',
                  'allow_no_indices', 'analyze_wildcard', 'analyzer', 'default_operator',
                  'df', 'expand_wildcards', 'explain', 'fielddata_fields', 'fields',
                  'from_', 'ignore_unavailable', 'lenient', 'lowercase_expanded_terms',
                  'preference', 'q', 'request_cache', 'routing', 'scroll', 'search_type',
                  'size', 'sort', 'stats', 'suggest_field', 'suggest_mode',
                  'suggest_size', 'suggest_text', 'terminate_after', 'timeout',
                  'track_scores', 'version')
    def count(self, index=None, doc_type=None, body=None, params=None, headers=None):
        searchable_indexes = self._normalize_index_to_list(index)

        i = 0
        for searchable_index in searchable_indexes:
            for document in self.__documents_dict[searchable_index]:
                if doc_type is not None and document.get('_type') != doc_type:
                    continue
                i += 1
        result = {
            'count': i,
            '_shards': {
                'successful': 1,
                'failed': 0,
                'total': 1
            }
        }

        return result

    @query_params('_source', '_source_exclude', '_source_include',
                  'allow_no_indices', 'analyze_wildcard', 'analyzer', 'default_operator',
                  'df', 'expand_wildcards', 'explain', 'fielddata_fields', 'fields',
                  'from_', 'ignore_unavailable', 'lenient', 'lowercase_expanded_terms',
                  'preference', 'q', 'request_cache', 'routing', 'scroll', 'search_type',
                  'size', 'sort', 'stats', 'suggest_field', 'suggest_mode',
                  'suggest_size', 'suggest_text', 'terminate_after', 'timeout',
                  'track_scores', 'version')
    def search(self, index=None, doc_type=None, body=None, params=None, headers=None):
        searchable_indexes = self._normalize_index_to_list(index)

        matches = []
        for searchable_index in searchable_indexes:
            for document in self.__documents_dict[searchable_index]:
                if doc_type:
                    if isinstance(doc_type, list) and document.get('_type') not in doc_type:
                        continue
                    if isinstance(doc_type, str) and document.get('_type') != doc_type:
                        continue
                matches.append(document)

        result = {
            'hits': {
                'total': len(matches),
                'max_score': 1.0
            },
            '_shards': {
                # Simulate indexes with 1 shard each
                'successful': len(searchable_indexes),
                'failed': 0,
                'total': len(searchable_indexes)
            },
            'took': 1,
            'timed_out': False
        }

        hits = []
        for match in matches:
            match['_score'] = 1.0
            hits.append(match)

        # build aggregations
        if body is not None and 'aggs' in body:
            aggregations = {}

            for aggregation, definition in body['aggs'].items():
                aggregations[aggregation] = {
                    "doc_count_error_upper_bound": 0,
                    "sum_other_doc_count": 0,
                    "buckets": []
                }

            if aggregations:
                result['aggregations'] = aggregations

        if 'scroll' in params:
            result['_scroll_id'] = str(get_random_scroll_id())
            params['size'] = int(params.get('size') if 'size' in params else 10)
            params['from'] = int(params.get('from') + params.get('size') if 'from' in params else 0)
            self.__scrolls[result.get('_scroll_id')] = {
                'index': index,
                'doc_type': doc_type,
                'body': body,
                'params': params
            }
            hits = hits[params.get('from'):params.get('from') + params.get('size')]

        result['hits']['hits'] = hits

        return result

    @query_params('scroll')
    def scroll(self, scroll_id, params=None, headers=None):
        scroll = self.__scrolls.pop(scroll_id)
        result = self.search(
            index=scroll.get('index'),
            doc_type=scroll.get('doc_type'),
            body=scroll.get('body'),
            params=scroll.get('params')
        )
        return result

    @query_params('consistency', 'parent', 'refresh', 'replication', 'routing',
                  'timeout', 'version', 'version_type')
    def delete(self, index, doc_type, id, params=None, headers=None):

        found = False

        if index in self.__documents_dict:
            for document in self.__documents_dict[index]:
                if document.get('_type') == doc_type and document.get('_id') == id:
                    found = True
                    self.__documents_dict[index].remove(document)
                    break

        result_dict = {
            'found': found,
            '_index': index,
            '_type': doc_type,
            '_id': id,
            '_version': 1,
        }

        if found:
            return result_dict
        else:
            raise NotFoundError(404, json.dumps(result_dict))

    @query_params('allow_no_indices', 'expand_wildcards', 'ignore_unavailable',
                  'preference', 'routing')
    def suggest(self, body, index=None, params=None, headers=None):
        if index is not None and index not in self.__documents_dict:
            raise NotFoundError(404, 'IndexMissingException[[{0}] missing]'.format(index))

        result_dict = {}
        for key, value in body.items():
            text = value.get('text')
            suggestion = int(text) + 1 if isinstance(text, int) else '{0}_suggestion'.format(text)
            result_dict[key] = [
                {
                    'text': text,
                    'length': 1,
                    'options': [
                        {
                            'text': suggestion,
                            'freq': 1,
                            'score': 1.0
                        }
                    ],
                    'offset': 0
                }
            ]
        return result_dict

    def _normalize_index_to_list(self, index):
        # Ensure to have a list of index
        if index is None:
            searchable_indexes = self.__documents_dict.keys()
        elif isinstance(index, str) or isinstance(index, unicode):
            searchable_indexes = [index]
        elif isinstance(index, list):
            searchable_indexes = index
        else:
            # Is it the correct exception to use ?
            raise ValueError("Invalid param 'index'")

        # Check index(es) exists
        for searchable_index in searchable_indexes:
            if searchable_index not in self.__documents_dict:
                raise NotFoundError(404, 'IndexMissingException[[{0}] missing]'.format(searchable_index))

        return searchable_indexes



from distutils.core import setup
setup(
  name = 'elasticmock_additional_apis',
  packages = ['elasticmock_additional_apis'],
  version = '0.4',
  license='MIT',
  description = 'Extended elasticmock module with additional Elasticsearch index APIs',
  author = 'Tereza Gabrielova',
  author_email = 'terezagabrielova@yahoo.com',
  url = 'https://github.com/Tessg22/elasticmock_additional_apis',
  download_url = 'https://github.com/Tessg22/elasticmock_additional_apis/archive/v_01.tar.gz',
  keywords = ['Tagsunittest', 'mock', 'elasticsearch'],
  install_requires=[
          'elasticsearch',
          'elasticmock',
          'functools',
          'mock',
          'random',
          'string',
      ],
  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
  ],
)



# -*- coding: utf-8 -*-

from elasticsearch.client.indices import IndicesClient
from elasticsearch.client.utils import query_params


class FakeIndicesClient(IndicesClient):

    @query_params('master_timeout', 'timeout')
    def create(self, index, body=None, params=None, headers=None):
        documents_dict = self.__get_documents_dict()
        if index not in documents_dict:
            documents_dict[index] = []

    @query_params('allow_no_indices', 'expand_wildcards', 'ignore_unavailable',
                  'local')
    def exists(self, index, params=None, headers=None):
        return index in self.__get_documents_dict()

    @query_params('allow_no_indices', 'expand_wildcards', 'force',
                  'ignore_unavailable', 'operation_threading')
    def refresh(self, index=None, params=None, headers=None):
        pass

    @query_params('master_timeout', 'timeout')
    def delete(self, index, params=None, headers=None):
        documents_dict = self.__get_documents_dict()
        if index in documents_dict:
            del documents_dict[index]

    def get_alias(self, prefix=None, params=None, headers=None):
        if prefix == "infra-*":
            return {
                        "infra-xyz-13-11-2020" : {
                                "filter" : {
                                    "term" : {
                                        "year" : 2020
                                    }
                            }
                        },
                        "infra-xyz-20-10-2021" : {
                                "filter" : {
                                    "term" : {
                                        "year" : 2021
                                    }
                            }
                        }
                    }

    def stats(self, index, params=None, headers=None):
        return {
                "uuid": "BLuIqmgBQyaWjfWuQUnX7Q",
                "primaries":
                {
                "docs": { "count": 12759052, "deleted": 0 },
                "store": { "size_in_bytes": 7660909325 },
                "indexing":
                    {
                    "index_total": 12759185,
                    "index_time_in_millis": 14299525,
                    "index_current": 0,
                    "index_failed": 0,
                    "delete_total": 0,
                    "delete_time_in_millis": 0,
                    "delete_current": 0,
                    "noop_update_total": 0,
                    "is_throttled": False,
                    "throttle_time_in_millis": 0,
                    },
                "get":
                    {
                    "total": 0,
                    "time_in_millis": 0,
                    "exists_total": 0,
                    "exists_time_in_millis": 0,
                    "missing_total": 0,
                    "missing_time_in_millis": 0,
                    "current": 0,
                    },
                "search":
                    {
                    "open_contexts": 0,
                    "query_total": 0,
                    "query_time_in_millis": 0,
                    "query_current": 0,
                    "fetch_total": 0,
                    "fetch_time_in_millis": 0,
                    "fetch_current": 0,
                    "scroll_total": 0,
                    "scroll_time_in_millis": 0,
                    "scroll_current": 0,
                    "suggest_total": 0,
                    "suggest_time_in_millis": 0,
                    "suggest_current": 0,
                    },
                "merges":
                    {
                    "current": 0,
                    "current_docs": 0,
                    "current_size_in_bytes": 0,
                    "total": 725,
                    "total_time_in_millis": 2568966,
                    "total_docs": 27087860,
                    "total_size_in_bytes": 16955534376,
                    "total_stopped_time_in_millis": 29857,
                    "total_throttled_time_in_millis": 1805866,
                    "total_auto_throttle_in_bytes": 15728640,
                    },
                "refresh": { "total": 6857, "total_time_in_millis": 421864, "listeners": 0 },
                "flush": { "total": 93, "periodic": 93, "total_time_in_millis": 20599 },
                "warmer": { "current": 0, "total": 6753, "total_time_in_millis": 4228 },
                "query_cache":
                    {
                    "memory_size_in_bytes": 0,
                    "total_count": 0,
                    "hit_count": 0,
                    "miss_count": 0,
                    "cache_size": 0,
                    "cache_count": 0,
                    "evictions": 0,
                    },
                "fielddata": { "memory_size_in_bytes": 0, "evictions": 0 },
                "completion": { "size_in_bytes": 0 },
                "segments":
                    {
                    "count": 74,
                    "memory_in_bytes": 14703764,
                    "terms_memory_in_bytes": 10316907,
                    "stored_fields_memory_in_bytes": 3987208,
                    "term_vectors_memory_in_bytes": 0,
                    "norms_memory_in_bytes": 122624,
                    "points_memory_in_bytes": 256145,
                    "doc_values_memory_in_bytes": 20880,
                    "index_writer_memory_in_bytes": 2039828,
                    "version_map_memory_in_bytes": 20083,
                    "fixed_bit_set_memory_in_bytes": 1598168,
                    "max_unsafe_auto_id_timestamp": -1,
                    "file_sizes": {},
                    },
                "translog":
                    {
                    "operations": 862217,
                    "size_in_bytes": 1705567621,
                    "uncommitted_operations": 45429,
                    "uncommitted_size_in_bytes": 89830471,
                    "earliest_last_modified_age": 0,
                    },
                "request_cache":
                    {
                    "memory_size_in_bytes": 0,
                    "evictions": 0,
                    "hit_count": 0,
                    "miss_count": 0,
                    },
                "recovery":
                    {
                    "current_as_source": 0,
                    "current_as_target": 0,
                    "throttle_time_in_millis": 0,
                    },
                },
                "total":
                {
                "docs": { "count": 25507574, "deleted": 0 },
                "store": { "size_in_bytes": 15521849910 },
                "indexing":
                    {
                    "index_total": 25507892,
                    "index_time_in_millis": 28577184,
                    "index_current": 2,
                    "index_failed": 0,
                    "delete_total": 0,
                    "delete_time_in_millis": 0,
                    "delete_current": 0,
                    "noop_update_total": 0,
                    "is_throttled": False,
                    "throttle_time_in_millis": 0,
                    },
                "get":
                    {
                    "total": 0,
                    "time_in_millis": 0,
                    "exists_total": 0,
                    "exists_time_in_millis": 0,
                    "missing_total": 0,
                    "missing_time_in_millis": 0,
                    "current": 0,
                    },
                "search":
                    {
                    "open_contexts": 0,
                    "query_total": 0,
                    "query_time_in_millis": 0,
                    "query_current": 0,
                    "fetch_total": 0,
                    "fetch_time_in_millis": 0,
                    "fetch_current": 0,
                    "scroll_total": 0,
                    "scroll_time_in_millis": 0,
                    "scroll_current": 0,
                    "suggest_total": 0,
                    "suggest_time_in_millis": 0,
                    "suggest_current": 0,
                    },
                "merges":
                    {
                    "current": 1,
                    "current_docs": 547828,
                    "current_size_in_bytes": 332108527,
                    "total": 1453,
                    "total_time_in_millis": 5151344,
                    "total_docs": 54474787,
                    "total_size_in_bytes": 34097980314,
                    "total_stopped_time_in_millis": 51270,
                    "total_throttled_time_in_millis": 3619556,
                    "total_auto_throttle_in_bytes": 31457280,
                    },
                "refresh":
                    { "total": 13740, "total_time_in_millis": 839446, "listeners": 0 },
                "flush": { "total": 187, "periodic": 187, "total_time_in_millis": 39921 },
                "warmer": { "current": 0, "total": 13534, "total_time_in_millis": 8610 },
                "query_cache":
                    {
                    "memory_size_in_bytes": 0,
                    "total_count": 0,
                    "hit_count": 0,
                    "miss_count": 0,
                    "cache_size": 0,
                    "cache_count": 0,
                    "evictions": 0,
                    },
                "fielddata": { "memory_size_in_bytes": 0, "evictions": 0 },
                "completion": { "size_in_bytes": 0 },
                "segments":
                    {
                    "count": 159,
                    "memory_in_bytes": 29491990,
                    "terms_memory_in_bytes": 20699814,
                    "stored_fields_memory_in_bytes": 7976816,
                    "term_vectors_memory_in_bytes": 0,
                    "norms_memory_in_bytes": 251840,
                    "points_memory_in_bytes": 512148,
                    "doc_values_memory_in_bytes": 51372,
                    "index_writer_memory_in_bytes": 6063988,
                    "version_map_memory_in_bytes": 48018,
                    "fixed_bit_set_memory_in_bytes": 3195472,
                    "max_unsafe_auto_id_timestamp": -1,
                    "file_sizes": {},
                    },
                "translog":
                    {
                    "operations": 1707331,
                    "size_in_bytes": 3377273219,
                    "uncommitted_operations": 142315,
                    "uncommitted_size_in_bytes": 281485673,
                    "earliest_last_modified_age": 0,
                    },
                "request_cache":
                    {
                    "memory_size_in_bytes": 0,
                    "evictions": 0,
                    "hit_count": 0,
                    "miss_count": 0,
                    },
                "recovery":
                    {
                    "current_as_source": 0,
                    "current_as_target": 0,
                    "throttle_time_in_millis": 0,
                    },
                },
                }

    def get(self, index, params=None, headers=None):
        return {
            index: {
                "aliases": {},
                "mappings": {},
                "settings": {
                    "index": {
                        "creation_date": "1429308615170",
                        "number_of_replicas": "1",
                        "number_of_shards": "5",
                        "uuid": "C5sqwXClSFyd5uF3MSrVgg",
                        "version": {
                        "created": "1050199"
                        }
                    }
                },
                "warmers": {}
            }
        }

    def __get_documents_dict(self):
        return self.client._FakeElasticsearch__documents_dict
